import math
from urllib.parse import urlparse

from loguru import logger
from requests import RequestException

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event

logger = logger.bind(name='next_sonarr_episodes')


class NextSonarrEpisodes:
    """
    This plugin return the 1st missing episode of every show configures in Sonarr.
    This can be used with the discover plugin or set_series_begin plugin to
    get the relevant data from Sonarr.

    Syntax:

    next_sonarr_episodes:
      base_url=<value> (Required)
      port=<value> (Default is 80)
      api_key=<value> (Required)
      include_ended=<yes|no> (Default is yes)
      only_monitored=<yes|no> (Default is yes)
      page_size=<value> (Default is 50)

    Page size determines the amount of results per each API call.
    Higher value means a bigger response. Lower value means more calls.
    Should be changed if there are performance issues.


    Usage: (Example with discover)

    discover_from_sonarr_task:
      discover:
        what:
          - next_sonarr_episodes:
              base_url: '{? credentials.sonarr.url ?}'
              port: 8989
              api_key: '{? credentials.sonarr.api_key ?}'
              include_ended: false
        from:
          - kat:
              verified: yes
      all_series: yes
      download: c:\bla\

    Usage: (Example with set_series_begin)

    set-series-begin-from-sonarr:
      next_sonarr_episodes:
        base_url: '{{ secrets.credentials.sonarr.url }}'
        port: 8989
        api_key: '{{ secrets.credentials.sonarr.api_key }}'
        include_ended: false
      accept_all: yes
      set_series_begin: yes
    """

    schema = {
        'type': 'object',
        'properties': {
            'base_url': {'type': 'string'},
            'port': {'type': 'number', 'default': 80},
            'api_key': {'type': 'string'},
            'include_ended': {'type': 'boolean', 'default': True},
            'only_monitored': {'type': 'boolean', 'default': True},
            'page_size': {'type': 'number', 'default': 50},
        },
        'required': ['api_key', 'base_url'],
        'additionalProperties': False,
    }

    # Function that gets a page number and page size and returns the responding result json
    def get_page(self, task, config, page_number):
        parsedurl = urlparse(config.get('base_url'))
        url = '{}://{}:{}{}/api/wanted/missing?page={}&pageSize={}&sortKey=series.title&sortdir=asc'.format(
            parsedurl.scheme,
            parsedurl.netloc,
            config.get('port'),
            parsedurl.path,
            page_number,
            config.get('page_size'),
        )
        headers = {'X-Api-Key': config['api_key']}
        try:
            json = task.requests.get(url, headers=headers).json()
        except RequestException as e:
            raise plugin.PluginError(
                'Unable to connect to Sonarr at {}://{}:{}{}. Error: {}'.format(
                    parsedurl.scheme, parsedurl.netloc, config.get('port'), parsedurl.path, e
                )
            )
        return json

    def on_task_input(self, task, config):
        json = self.get_page(task, config, 1)
        pages = int(
            math.ceil(json['totalRecords'] / config.get('page_size'))
        )  # Sets number of requested pages
        current_series_id = 0  # Initializes current series parameter
        for page in range(1, pages + 1):
            json = self.get_page(task, config, page)
            for record in json['records']:
                # Verifies that we only get the first missing episode from a series
                if current_series_id != record['seriesId']:
                    current_series_id = record['seriesId']
                    season = record['seasonNumber']
                    episode = record['episodeNumber']
                    entry = Entry(
                        url='',
                        series_name=record['series']['title'],
                        series_season=season,
                        series_episode=episode,
                        series_id='S%02dE%02d' % (season, episode),
                        tvdb_id=record['series'].get('tvdbId'),
                        tvrage_id=record['series'].get('tvRageId'),
                        tvmaze_id=record['series'].get('tvMazeId'),
                        title=record['series']['title'] + ' ' + 'S%02dE%02d' % (season, episode),
                    )
                    # Test mode logging
                    if entry and task.options.test:
                        logger.verbose("Test mode. Entry includes:")
                        for key, value in list(entry.items()):
                            logger.verbose('     {}: {}', key.capitalize(), value)
                    if entry.isvalid():
                        yield entry
                    else:
                        logger.error('Invalid entry created? {}', entry)


@event('plugin.register')
def register_plugin():
    plugin.register(NextSonarrEpisodes, 'next_sonarr_episodes', api_ver=2)



from datetime import datetime

from loguru import logger
from sqlalchemy import Column, DateTime, Integer, String, Unicode

from flexget import db_schema, plugin
from flexget.db_schema import Session
from flexget.entry import Entry
from flexget.event import event
from flexget.utils import qualities
from flexget.utils.database import quality_property
from flexget.utils.tools import group_entries, parse_timedelta

logger = logger.bind(name='upgrade')

Base = db_schema.versioned_base('upgrade', 0)

entry_actions = {'accept': Entry.accept, 'reject': Entry.reject, 'fail': Entry.fail}


class EntryUpgrade(Base):
    __tablename__ = 'upgrade'

    id = Column(Unicode, primary_key=True, index=True)
    title = Column(Unicode)
    _quality = Column('quality', String)
    quality = quality_property('_quality')
    proper_count = Column(Integer, default=0)
    first_seen = Column(DateTime, default=datetime.now)
    updated = Column(DateTime, index=True, default=datetime.now)

    def __str__(self):
        return '<Upgrade(id=%s,added=%s,quality=%s)>' % (self.id, self.added, self.quality)


class FilterUpgrade:
    schema = {
        'type': 'object',
        'properties': {
            'identified_by': {'type': 'string'},
            'tracking': {'type': 'boolean'},
            'target': {'type': 'string', 'format': 'quality_requirements'},
            'on_lower': {'type': 'string', 'enum': ['accept', 'reject', 'do_nothing']},
            'timeframe': {'type': 'string', 'format': 'interval'},
            'propers': {'type': 'boolean'},
        },
        'additionalProperties': False,
    }

    def prepare_config(self, config):
        if not config or config is False:
            return

        if config is True:
            config = {}

        config.setdefault('identified_by', 'auto')
        config.setdefault('tracking', True)
        config.setdefault('on_lower', 'do_nothing')
        config.setdefault('target', None)
        config.setdefault('timeframe', None)
        config.setdefault('propers', True)
        return config

    def filter_entries(self, entries, existing, target, action_on_lower):

        target_requirement = qualities.Requirements(target) if target else None
        filtered = []

        for entry in entries:
            # Filter out entries within target
            if target:
                if not target_requirement.allows(entry['quality']):
                    logger.debug(
                        'Skipping {} as does not meet upgrade quality requirements', entry['title']
                    )
                    if action_on_lower:
                        action_on_lower(entry, 'does not meet upgrade quality requirements')
                    continue

            if entry['quality'] < existing.quality:
                logger.debug('Skipping {} as lower quality then existing', entry['title'])
                if action_on_lower:
                    action_on_lower(entry, 'lower quality then existing')
                continue

            if (
                entry['quality'] == existing.quality
                and entry.get('proper_count', 0) <= existing.proper_count
            ):
                logger.debug('Skipping {} as same quality but lower proper', entry['title'])
                if action_on_lower:
                    action_on_lower(entry, 'lower proper then existing')
                continue

            filtered.append(entry)

        return filtered

    def on_task_filter(self, task, config):
        config = self.prepare_config(config)

        if not config or not config['target']:
            return

        identified_by = (
            '{{ media_id }}' if config['identified_by'] == 'auto' else config['identified_by']
        )

        grouped_entries = group_entries(task.accepted + task.undecided, identified_by)
        if not grouped_entries:
            return

        with Session() as session:
            # Prefetch Data
            existing_ids = (
                session.query(EntryUpgrade)
                .filter(EntryUpgrade.id.in_(grouped_entries.keys()))
                .all()
            )
            existing_ids = {e.id: e for e in existing_ids}

            for identifier, entries in grouped_entries.items():
                if not entries:
                    continue

                existing = existing_ids.get(identifier)
                if not existing:
                    # No existing, do_nothing
                    continue

                logger.debug(
                    'Looking for upgrades for identifier {} (within {} entries)',
                    identifier,
                    len(entries),
                )

                # Check if passed allowed timeframe
                if config['timeframe']:
                    expires = existing.first_seen + parse_timedelta(config['timeframe'])
                    if expires <= datetime.now():
                        # Timeframe reached, allow
                        logger.debug(
                            'Skipping upgrade with identifier {} as timeframe reached', identifier
                        )
                        continue

                # Filter out lower quality and propers
                action_on_lower = (
                    entry_actions[config['on_lower']]
                    if config['on_lower'] != 'do_nothing'
                    else None
                )
                upgradeable = self.filter_entries(
                    entries, existing, config['target'], action_on_lower
                )

                # Skip if we have no entries after filtering
                if not upgradeable:
                    continue

                # Sort entities in order of quality and best proper
                upgradeable.sort(
                    key=lambda e: (e['quality'], e.get('proper_count', 0)), reverse=True
                )

                # First entry will be the best quality
                best = upgradeable.pop(0)
                best.accept('upgraded quality')
                logger.debug(
                    'Found {} as upgraded quality for identifier {}', best['title'], identifier
                )

                # Process rest
                for entry in upgradeable:
                    logger.debug(
                        'Skipping {} as lower quality then best {}', entry['title'], best['title']
                    )
                    if action_on_lower:
                        action_on_lower(entry, 'lower quality then best match')

    def on_task_learn(self, task, config):
        config = self.prepare_config(config)
        if not config or not config['tracking']:
            return

        identified_by = (
            '{{ media_id }}' if config['identified_by'] == 'auto' else config['identified_by']
        )

        grouped_entries = group_entries(task.accepted, identified_by)
        if not grouped_entries:
            return

        with Session() as session:
            # Prefetch Data
            existing_ids = (
                session.query(EntryUpgrade)
                .filter(EntryUpgrade.id.in_(grouped_entries.keys()))
                .all()
            )
            existing_ids = {e.id: e for e in existing_ids}

            for identifier, entries in grouped_entries.items():
                if not entries:
                    continue

                # Sort entities in order of quality
                entries.sort(key=lambda e: e['quality'], reverse=True)
                # First entry will be the best quality
                best_entry = entries[0]

                existing = existing_ids.get(identifier)

                if not existing:
                    existing = EntryUpgrade()
                    existing.id = identifier
                    session.add(existing)
                elif existing.quality > best_entry['quality']:
                    continue

                existing.quality = best_entry['quality']
                existing.title = best_entry['title']
                existing.proper_count = best_entry.get('proper_count', 0)
                existing.updated = datetime.now()

                logger.debug(
                    'Tracking upgrade on identifier `{}` current quality `{}`',
                    identifier,
                    best_entry['quality'],
                )


@event('plugin.register')
def register_plugin():
    plugin.register(FilterUpgrade, 'upgrade', api_ver=2)



from flexget import options, plugin
from flexget.event import event
from flexget.manager import Session
from flexget.terminal import TerminalTable, TerminalTableError, console, table_parser

from . import db


def action_auth(options):
    if not (options.account):
        console(
            'You must specify an account (local identifier) so we know where to save your access token!'
        )
        return
    try:
        db.get_access_token(options.account, options.pin, re_auth=True, called_from_cli=True)
        console('Successfully authorized Flexget app on Trakt.tv. Enjoy!')
        return
    except plugin.PluginError as e:
        console('Authorization failed: %s' % e)


def action_list(options):
    with Session() as session:
        if not options.account:
            # Print all accounts
            accounts = session.query(db.TraktUserAuth).all()
            if not accounts:
                console('No trakt authorizations stored in database.')
                return
            header = ['Account', 'Created', 'Expires']
            table_data = [header]

            for auth in accounts:
                table_data.append(
                    [
                        auth.account,
                        auth.created.strftime('%Y-%m-%d'),
                        auth.expires.strftime('%Y-%m-%d'),
                    ]
                )
            try:
                table = TerminalTable(options.table_type, table_data)
                console(table.output)
                return
            except TerminalTableError as e:
                console('ERROR: %s' % str(e))

        # Show a specific account
        acc = (
            session.query(db.TraktUserAuth)
            .filter(db.TraktUserAuth.account == options.account)
            .first()
        )
        if acc:
            console('Authorization expires on %s' % acc.expires)
        else:
            console('Flexget has not been authorized to access your account.')


def action_refresh(options):
    if not options.account:
        console('Please specify an account')
        return
    try:
        db.get_access_token(options.account, refresh=True)
        console('Successfully refreshed your access token.')
        return
    except plugin.PluginError as e:
        console('Authorization failed: %s' % e)


def action_delete(options):
    if not options.account:
        console('Please specify an account')
        return
    try:
        db.delete_account(options.account)
        console('Successfully deleted your access token.')
        return
    except plugin.PluginError as e:
        console('Deletion failed: %s' % e)


def do_cli(manager, options):
    action_map = {
        'auth': action_auth,
        'list': action_list,
        'refresh': action_refresh,
        'delete': action_delete,
    }

    action_map[options.action](options)


@event('options.register')
def register_parser_arguments():
    acc_text = 'Local identifier which should be used in your config to refer these credentials'
    # Register subcommand
    parser = options.register_command(
        'trakt', do_cli, help='View and manage trakt authentication.'
    )
    # Set up our subparsers
    subparsers = parser.add_subparsers(title='actions', metavar='<action>', dest='action')
    auth_parser = subparsers.add_parser(
        'auth', help='Authorize Flexget to access your Trakt.tv account'
    )

    auth_parser.add_argument('account', metavar='<account>', help=acc_text)
    auth_parser.add_argument(
        'pin',
        metavar='<pin>',
        help='Get this by authorizing FlexGet to use your trakt account '
        'at %s. WARNING: DEPRECATED.' % db.PIN_URL,
        nargs='?',
    )

    show_parser = subparsers.add_parser(
        'list',
        help='List expiration date for Flexget authorization(s) (don\'t worry, '
        'they will automatically refresh when expired)',
        parents=[table_parser],
    )
    show_parser.add_argument('account', metavar='<account>', nargs='?', help=acc_text)

    refresh_parser = subparsers.add_parser(
        'refresh',
        help='Manually refresh your access token associated with your' ' --account <name>',
    )
    refresh_parser.add_argument('account', metavar='<account>', help=acc_text)

    delete_parser = subparsers.add_parser(
        'delete', help='Delete the specified <account> name from local database'
    )
    delete_parser.add_argument('account', metavar='<account>', help=acc_text)
